/*******************************************************************************
 * Copyright (c) 2013 IBM Corp.
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * and Eclipse Distribution License v1.0 which accompany this distribution.
 *
 * The Eclipse Public License is available at
 *    http://www.eclipse.org/legal/epl-v10.html
 * and the Eclipse Distribution License is available at
 *   http://www.eclipse.org/org/documents/edl-v10.php.
 *
 * Contributors:
 *    Takahiro Inaba - initial API and implementation and/or initial
 *                     documentation
 *******************************************************************************/
package com.ibm.jmeter.protocol.mqtt;

import static org.junit.Assert.*;

import java.lang.reflect.Field;
import java.nio.charset.Charset;

import org.apache.commons.codec.binary.Base64;
import org.apache.jmeter.samplers.SampleResult;
import org.junit.After;
import org.junit.Test;

public class MqttSamplerTest {

	@After
	public void removeThreadLocalValue() throws Exception {
		Field f = MqttSampler.class.getDeclaredField("tlClientCache");
		f.setAccessible(true);
		((ThreadLocal) f.get(MqttSampler.class)).remove();
	}

	@Test
	public void testStandard() {
		TestMqttSampler sampler0 = new TestMqttSampler();

		TestMqttSampler sampler1 = new TestMqttSampler();
		sampler1.setServerURI("tcp://localhost:1883");
		sampler1.setClientId("Client01");
		sampler1.setReuseConnection(true);
		sampler1.setCloseConnection(false);
		sampler1.setTopicString("Topic01");
		sampler1.setQos(0);
		sampler1.setRetained(false);
		sampler1.setBase64Encoded(false);
		sampler1.setMessageBody("1");

		TestMqttSampler sampler2 = new TestMqttSampler();
		sampler2.setServerURI("tcp://localhost:1883");
		sampler2.setClientId("Client01");
		sampler2.setReuseConnection(true);
		sampler2.setCloseConnection(false);
		sampler2.setTopicString("Topic01");
		sampler2.setQos(0);
		sampler2.setRetained(false);
		sampler2.setBase64Encoded(false);
		sampler2.setMessageBody("2");

		sampler0.testStarted();

		SampleResult result1 = sampler1.sample(null);
		assertTrue(result1.isSuccessful());
		MockClient client1 = sampler1.client;
		assertTrue(client1.hadBeenConnected);
		assertTrue(client1.connected);
		assertArrayEquals("1".getBytes(), ((MqttMessage) client1.givenMessages.get(0)).getMessageBody());
		assertEquals("Topic01", client1.givenTopicStrings.get(0));
		assertEquals(client1.numOfPublish, 1);

		SampleResult result2 = sampler2.sample(null);
		assertTrue(result2.isSuccessful());
		assertNull(sampler2.client);
		assertTrue(client1.hadBeenConnected);
		assertTrue(client1.connected);
		assertArrayEquals("2".getBytes(), ((MqttMessage) client1.givenMessages.get(1)).getMessageBody());
		assertEquals("Topic01", client1.givenTopicStrings.get(1));
		assertEquals(client1.numOfPublish, 2);

		sampler0.testEnded();
		assertFalse(client1.connected);
	}

	@Test
	public void testDoNotReuseConnection() {
		TestMqttSampler sampler0 = new TestMqttSampler();

		TestMqttSampler sampler1 = new TestMqttSampler();
		sampler1.setServerURI("tcp://localhost:1883");
		sampler1.setClientId("Client01");
		sampler1.setReuseConnection(false);
		sampler1.setCloseConnection(false);
		sampler1.setTopicString("Topic01");
		sampler1.setQos(0);
		sampler1.setRetained(false);
		sampler1.setBase64Encoded(false);
		sampler1.setMessageBody("1");

		TestMqttSampler sampler2 = new TestMqttSampler();
		sampler2.setServerURI("tcp://localhost:1883");
		sampler2.setClientId("Client01");
		sampler2.setReuseConnection(false);
		sampler2.setCloseConnection(false);
		sampler2.setTopicString("Topic01");
		sampler2.setQos(0);
		sampler2.setRetained(false);
		sampler2.setBase64Encoded(false);
		sampler2.setMessageBody("2");

		sampler0.testStarted();

		SampleResult result1 = sampler1.sample(null);
		assertTrue(result1.isSuccessful());
		MockClient client1 = sampler1.client;
		assertTrue(client1.hadBeenConnected);
		assertTrue(client1.connected);
		assertArrayEquals("1".getBytes(), ((MqttMessage) client1.givenMessages.get(0)).getMessageBody());
		assertEquals("Topic01", client1.givenTopicStrings.get(0));
		assertEquals(client1.numOfPublish, 1);


		SampleResult result2 = sampler2.sample(null);
		assertTrue(result2.isSuccessful());
		MockClient client2 = sampler2.client;
		assertTrue(client2.hadBeenConnected);
		assertTrue(client2.connected);
		assertArrayEquals("2".getBytes(), ((MqttMessage) client2.givenMessages.get(0)).getMessageBody());
		assertEquals("Topic01", client2.givenTopicStrings.get(0));
		assertEquals(client2.numOfPublish, 1);

		sampler0.testEnded();
		assertFalse(client1.connected);
		assertFalse(client2.connected);
	}

	@Test
	public void testCloseConnection() {
		TestMqttSampler sampler0 = new TestMqttSampler();

		TestMqttSampler sampler1 = new TestMqttSampler();
		sampler1.setServerURI("tcp://localhost:1883");
		sampler1.setClientId("Client01");
		sampler1.setReuseConnection(true);
		sampler1.setCloseConnection(true);
		sampler1.setTopicString("Topic01");
		sampler1.setQos(0);
		sampler1.setRetained(false);
		sampler1.setBase64Encoded(false);
		sampler1.setMessageBody("1");

		TestMqttSampler sampler2 = new TestMqttSampler();
		sampler2.setServerURI("tcp://localhost:1883");
		sampler2.setClientId("Client01");
		sampler2.setReuseConnection(true);
		sampler2.setCloseConnection(true);
		sampler2.setTopicString("Topic01");
		sampler2.setQos(0);
		sampler2.setRetained(false);
		sampler2.setBase64Encoded(false);
		sampler2.setMessageBody("2");

		sampler0.testStarted();

		SampleResult result1 = sampler1.sample(null);
		assertTrue(result1.isSuccessful());
		MockClient client1 = sampler1.client;
		assertTrue(client1.hadBeenConnected);
		assertFalse(client1.connected);
		assertArrayEquals("1".getBytes(), ((MqttMessage) client1.givenMessages.get(0)).getMessageBody());
		assertEquals("Topic01", client1.givenTopicStrings.get(0));
		assertEquals(client1.numOfPublish, 1);


		SampleResult result2 = sampler2.sample(null);
		assertTrue(result2.isSuccessful());
		MockClient client2 = sampler2.client;
		assertTrue(client2.hadBeenConnected);
		assertFalse(client2.connected);
		assertArrayEquals("2".getBytes(), ((MqttMessage) client2.givenMessages.get(0)).getMessageBody());
		assertEquals("Topic01", client2.givenTopicStrings.get(0));
		assertEquals(client2.numOfPublish, 1);

		sampler0.testEnded();
		assertFalse(client1.connected);
		assertFalse(client2.connected);
	}

	@Test
	public void testDoNotReuseAndCloseConnection() {
		TestMqttSampler sampler0 = new TestMqttSampler();

		TestMqttSampler sampler1 = new TestMqttSampler();
		sampler1.setServerURI("tcp://localhost:1883");
		sampler1.setClientId("Client01");
		sampler1.setReuseConnection(false);
		sampler1.setCloseConnection(true);
		sampler1.setTopicString("Topic01");
		sampler1.setQos(0);
		sampler1.setRetained(false);
		sampler1.setBase64Encoded(false);
		sampler1.setMessageBody("1");

		TestMqttSampler sampler2 = new TestMqttSampler();
		sampler2.setServerURI("tcp://localhost:1883");
		sampler2.setClientId("Client01");
		sampler2.setReuseConnection(false);
		sampler2.setCloseConnection(true);
		sampler2.setTopicString("Topic01");
		sampler2.setQos(0);
		sampler2.setRetained(false);
		sampler2.setBase64Encoded(false);
		sampler2.setMessageBody("2");

		sampler0.testStarted();

		SampleResult result1 = sampler1.sample(null);
		assertTrue(result1.isSuccessful());
		MockClient client1 = sampler1.client;
		assertTrue(client1.hadBeenConnected);
		assertFalse(client1.connected);
		assertArrayEquals("1".getBytes(), ((MqttMessage) client1.givenMessages.get(0)).getMessageBody());
		assertEquals("Topic01", client1.givenTopicStrings.get(0));
		assertEquals(client1.numOfPublish, 1);


		SampleResult result2 = sampler2.sample(null);
		assertTrue(result2.isSuccessful());
		MockClient client2 = sampler2.client;
		assertTrue(client2.hadBeenConnected);
		assertFalse(client2.connected);
		assertArrayEquals("2".getBytes(), ((MqttMessage) client2.givenMessages.get(0)).getMessageBody());
		assertEquals("Topic01", client2.givenTopicStrings.get(0));
		assertEquals(client2.numOfPublish, 1);

		sampler0.testEnded();
		assertFalse(client1.connected);
		assertFalse(client2.connected);
	}

	@Test
	public void testDifferentClientId() {
		TestMqttSampler sampler0 = new TestMqttSampler();

		TestMqttSampler sampler1 = new TestMqttSampler();
		sampler1.setServerURI("tcp://localhost:1883");
		sampler1.setClientId("Client01");
		sampler1.setReuseConnection(true);
		sampler1.setCloseConnection(false);
		sampler1.setTopicString("Topic01");
		sampler1.setQos(0);
		sampler1.setRetained(false);
		sampler1.setBase64Encoded(false);
		sampler1.setMessageBody("1");

		TestMqttSampler sampler2 = new TestMqttSampler();
		sampler2.setServerURI("tcp://localhost:1883");
		sampler2.setClientId("Client02");
		sampler2.setReuseConnection(true);
		sampler2.setCloseConnection(false);
		sampler2.setTopicString("Topic01");
		sampler2.setQos(0);
		sampler2.setRetained(false);
		sampler2.setBase64Encoded(false);
		sampler2.setMessageBody("2");

		sampler0.testStarted();

		SampleResult result1 = sampler1.sample(null);
		assertTrue(result1.isSuccessful());
		MockClient client1 = sampler1.client;
		assertTrue(client1.hadBeenConnected);
		assertTrue(client1.connected);
		assertArrayEquals("1".getBytes(), ((MqttMessage) client1.givenMessages.get(0)).getMessageBody());
		assertEquals("Topic01", client1.givenTopicStrings.get(0));
		assertEquals(client1.numOfPublish, 1);


		SampleResult result2 = sampler2.sample(null);
		assertTrue(result2.isSuccessful());
		MockClient client2 = sampler2.client;
		assertTrue(client2.hadBeenConnected);
		assertTrue(client2.connected);
		assertArrayEquals("2".getBytes(), ((MqttMessage) client2.givenMessages.get(0)).getMessageBody());
		assertEquals("Topic01", client2.givenTopicStrings.get(0));
		assertEquals(client2.numOfPublish, 1);

		sampler0.testEnded();
		assertFalse(client1.connected);
		assertFalse(client2.connected);
	}

	@Test
	public void testEncodedBody() {
		TestMqttSampler sampler = new TestMqttSampler();
		sampler.setServerURI("tcp://localhost:1883");
		sampler.setClientId("Client01");
		sampler.setBase64Encoded(true);
		sampler.setMessageBody(Base64.encodeBase64String("1".getBytes()));

		SampleResult result1 = sampler.sample(null);
		assertTrue(result1.isSuccessful());
		MockClient client1 = sampler.client;
		assertArrayEquals("1".getBytes(), ((MqttMessage) client1.givenMessages.get(0)).getMessageBody());
	}

	@Test
	public void testFailOnPublish() {
		TestMqttSampler sampler0 = new TestMqttSampler();

		TestMqttSampler sampler1 = new TestMqttSampler();
		sampler1.setServerURI("tcp://localhost:1883");
		sampler1.setClientId("Client01");
		sampler1.failOnPublish = true;
		sampler1.setMessageBody("1");

		sampler0.testStarted();

		SampleResult result1 = sampler1.sample(null);
		assertFalse(result1.isSuccessful());
		MockClient client1 = sampler1.client;

		sampler0.testEnded();
		assertFalse(client1.connected);
	}

	@Test
	public void testProtocols() {
		// TCP
		{
			TestMqttSampler sampler = new TestMqttSampler();
			sampler.setServerURI("tcp://localhost:1883");
			sampler.setClientId("Client01");
			sampler.setMessageBody("1");
			sampler.sample(null);
			assertTrue(sampler.realClient instanceof MqttOverTcpIpClient);
		}

		// SSL
		{
			TestMqttSampler sampler = new TestMqttSampler();
			sampler.setServerURI("ssl://localhost:8883");
			sampler.setClientId("Client01");
			sampler.setMessageBody("1");
			sampler.sample(null);
			assertTrue(sampler.realClient instanceof MqttOverTcpIpClient);
		}

		// Not supported (WebSocket)
		{
			TestMqttSampler sampler = new TestMqttSampler();
			sampler.setServerURI("ws://localhost:1883/mqtt");
			sampler.setClientId("Client01");
			sampler.setMessageBody("1");
			assertFalse(sampler.sample(null).isSuccessful());
		}
	}

	@Test
	public void testCharacterEncoding() {
		TestMqttSampler sampler = new TestMqttSampler();
		sampler.setServerURI("tcp://localhost:1883");
		sampler.setClientId("Client01");
		sampler.setMessageBody("日本語");	// "Japanese" in Japanese
		sampler.setCharacterEncoding("Windows-31J");

		SampleResult result1 = sampler.sample(null);
		assertTrue(result1.isSuccessful());
		MockClient client1 = sampler.client;
		assertArrayEquals("日本語".getBytes(Charset.forName("Windows-31J")),
				((MqttMessage) client1.givenMessages.get(0)).getMessageBody());
	}
}
